import json
import os.path

tags = {
    '1.0': '1__0__3',
    '1.1': '1__1__4',
}

schema_url = 'https://standard.open-contracting.org/schema/{}/release-schema.json'


def path(filename):
    return os.path.join('tests', 'fixtures', filename)


def read(filename, mode='rt', encoding=None, **kwargs):
    with open(path(filename), mode, encoding=encoding, **kwargs) as f:
        return f.read()


def load(*args, **kwargs):
    return json.loads(read(*args, **kwargs))



import json
import os.path

import pytest

from ocdsmerge import Merger


@pytest.fixture()
def vcr_cassette_name(request):
    return request.node.name.split('[', 1)[0]


@pytest.fixture()
def simple_merger():
    with open(os.path.join('tests', 'fixtures', 'schema.json')) as f:
        return Merger(json.load(f))


@pytest.fixture()
def empty_merger():
    return Merger({})



import re
from functools import lru_cache

import requests

from ocdsmerge.exceptions import (MissingDateKeyError, NonObjectReleaseError, NonStringDateValueError,
                                  NullDateValueError)


@lru_cache()
def get_tags():
    """
    Returns the tags of all versions of OCDS in alphabetical order.
    """
    return re.findall(r'"(\d+__\d+__\d+)/', requests.get('https://standard.open-contracting.org/schema/').text)


def get_release_schema_url(tag):
    """
    Returns the URL of the release schema in the given version of OCDS.
    """
    return 'https://standard.open-contracting.org/schema/{}/release-schema.json'.format(tag)


# If we need a method to get dates from releases, see https://github.com/open-contracting/ocds-merge/issues/25
def sorted_releases(releases):
    """
    Sorts a list of releases by date.
    """
    # Avoids an error if sorting a single compiled release.
    if isinstance(releases, list) and len(releases) == 1 and isinstance(releases[0], dict):
        return releases
    try:
        return sorted(releases, key=lambda release: release['date'])
    except KeyError:
        raise MissingDateKeyError('date', 'The `date` field of at least one release is missing.')
    except TypeError as e:
        if ' not supported between instances of ' in e.args[0]:
            if 'NoneType' in e.args[0]:
                raise NullDateValueError('The `date` field of at least one release is null.')
            else:
                raise NonStringDateValueError('The `date` field of at least one release is not a string.')
        elif e.args[0] == 'string indices must be integers':
            raise NonObjectReleaseError('At least one release is a string, not a dict. Use `json.loads` to parse the '
                                        'string as JSON.')
        elif e.args[0] == 'byte indices must be integers or slices, not str':
            raise NonObjectReleaseError('At least one release is a byte-string, not a dict. Use `json.loads` to parse '
                                        'the byte-string as JSON.')
        elif e.args[0] == 'list indices must be integers or slices, not str':
            raise NonObjectReleaseError('At least one release is a list, not a dict.')
        elif e.args[0] == 'tuple indices must be integers or slices, not str':
            raise NonObjectReleaseError('At least one release is a tuple, not a dict.')
        elif e.args[0] == "'set' object is not subscriptable":
            raise NonObjectReleaseError('At least one release is a set, not a dict.')
        else:
            raise



from functools import lru_cache

import jsonref

from ocdsmerge.util import get_release_schema_url, get_tags


def get_merge_rules(schema=None):
    """
    Returns merge rules as key-value pairs, in which the key is a JSON path as a tuple, and the value is a list of
    merge properties whose values are `true`.
    """
    schema = schema or get_release_schema_url(get_tags()[-1])
    if isinstance(schema, dict):
        return _get_merge_rules_from_dereferenced_schema(jsonref.JsonRef.replace_refs(schema))
    else:
        return _get_merge_rules_from_url_or_path(schema)


@lru_cache()
def _get_merge_rules_from_url_or_path(schema):
    if schema.startswith('http'):
        deref_schema = jsonref.load_uri(schema)
    else:
        with open(schema) as f:
            deref_schema = jsonref.load(f)
    return _get_merge_rules_from_dereferenced_schema(deref_schema)


def _get_merge_rules_from_dereferenced_schema(deref_schema):
    return dict(_get_merge_rules(deref_schema['properties']))


def _get_merge_rules(properties, path=None):
    """
    Yields merge rules as key-value pairs, in which the first element is a JSON path as a tuple, and the second element
    is a list of merge properties whose values are `true`.
    """
    if path is None:
        path = ()

    for key, value in properties.items():
        new_path = path + (key,)
        types = _get_types(value)

        # `omitWhenMerged` supersedes all other rules.
        # See https://standard.open-contracting.org/1.1/en/schema/merging/#discarded-fields
        if value.get('omitWhenMerged') or value.get('mergeStrategy') == 'ocdsOmit':
            yield new_path, {'omitWhenMerged'}
        # `wholeListMerge` supersedes any nested rules.
        # See https://standard.open-contracting.org/1.1/en/schema/merging/#whole-list-merge
        elif 'array' in types and (value.get('wholeListMerge') or value.get('mergeStrategy') == 'ocdsVersion'):
            yield new_path, {'wholeListMerge'}
        # See https://standard.open-contracting.org/1.1/en/schema/merging/#object-values
        elif 'object' in types and 'properties' in value:
            yield from _get_merge_rules(value['properties'], path=new_path)
        # See https://standard.open-contracting.org/1.1/en/schema/merging/#whole-list-merge
        elif 'array' in types and 'items' in value:
            item_types = _get_types(value['items'])
            if any(item_type != 'object' for item_type in item_types):
                yield new_path, {'wholeListMerge'}
            elif 'object' in item_types and 'properties' in value['items']:
                if 'id' not in value['items']['properties']:
                    yield new_path, {'wholeListMerge'}
                else:
                    yield from _get_merge_rules(value['items']['properties'], path=new_path)


def _get_types(prop):
    """
    Returns a property's `type` as a list.
    """
    if 'type' not in prop:
        return []
    if isinstance(prop['type'], str):
        return [prop['type']]
    return prop['type']



import uuid
import warnings
from enum import Enum, auto, unique

from ocdsmerge.exceptions import DuplicateIdValueWarning, InconsistentTypeError

VERSIONED_VALUE_KEYS = frozenset(['releaseID', 'releaseDate', 'releaseTag', 'value'])


@unique
class MergeStrategy(Enum):
    APPEND = auto()
    MERGE_BY_POSITION = auto()


globals().update(MergeStrategy.__members__)


class IdValue(str):
    """
    A string with ``identifier`` and ``original_value`` properties.
    """
    def __init__(self, identifier):
        self.identifier = identifier
        str.__init__(identifier)

    @property
    def original_value(self):
        return self._original_value

    @original_value.setter
    def original_value(self, original_value):
        self._original_value = original_value


class IdDict(dict):
    """
    A dictionary with an ``identifier`` property.
    """
    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, identifier):
        self._identifier = identifier


def is_versioned_value(value):
    """
    Returns whether the value is a versioned value.
    """
    return len(value) == 4 and VERSIONED_VALUE_KEYS.issuperset(value)


def flatten(obj, merge_rules, rule_overrides, flattened, path=(), rule_path=(), versioned=False):
    """
    Flattens a JSON object into key-value pairs, in which the key is the JSON path as a tuple. For example:

    Replaces numbers in JSON paths (representing positions in arrays) with special objects. This ensures that objects
    in arrays with different `id` values have different JSON paths – and makes it easy to identify such arrays.

    .. code:: json

       {
           "a": "I am a",
           "b": ["A", "list"],
           "c": [
               {"id": 1, "cb": "I am ca"},
               {"id": 2, "ca": "I am cb"}
           ]
       }

    flattens to:

    .. code:: python

       {
           ('a',): 'I am a',
           ('b',): ['A', 'list'],
           ('a', '1', 'cb'): 'I am ca',
           ('a', '1', 'id'): 1,
           ('a', '2', 'ca'): 'I am cb',
           ('a', '2', 'id'): 2,
       }
    """
    # For an exploration of alternatives, see: https://github.com/open-contracting/ocds-merge/issues/26

    if type(obj) is list:
        is_dict = False
        iterable = _enumerate(obj, path, rule_path, rule_overrides.get(rule_path))
        new_rule_path = rule_path
    else:
        is_dict = True
        iterable = obj.items()

    for key, value in iterable:
        if is_dict:
            new_rule_path = rule_path + (key,)

        new_path_merge_rules = merge_rules.get(new_rule_path, set())

        if 'omitWhenMerged' in new_path_merge_rules:
            continue
        # If it's `wholeListMerge`, if it's neither an object nor an array, if it's an array containing non-objects
        # (even if `wholeListMerge` is `false`), or if it's versioned values, use the whole list merge strategy.
        # Note: Behavior is undefined and inconsistent if the array is not in the schema and contains objects in some
        # cases but not in others.
        # See https://standard.open-contracting.org/1.1/en/schema/merging/#whole-list-merge
        # See https://standard.open-contracting.org/1.1/en/schema/merging/#objects
        elif 'wholeListMerge' in new_path_merge_rules or not isinstance(value, (dict, list)) or \
                type(value) is list and any(not isinstance(item, dict) for item in value) or \
                versioned and value and all(is_versioned_value(item) for item in value):
            flattened[path + (key,)] = value
        # Recurse into non-empty objects, and arrays of objects that aren't `wholeListMerge`.
        elif value:
            flatten(value, merge_rules, rule_overrides, flattened, path + (key,), new_rule_path, versioned)

    return flattened


def _enumerate(obj, path, rule_path, rule):
    # This tracks the identifiers of objects in an array, to warn about collisions.
    identifiers = {}

    for key, value in enumerate(obj):
        new_key, default_key = _id_value(key, value, rule)

        # Check whether the identifier is used by other objects in the array.
        default_path = path + (default_key,)
        if default_path not in identifiers:
            identifiers[default_path] = key
        elif identifiers[default_path] != key:
            warnings.warn(DuplicateIdValueWarning(rule_path, default_key, 'Multiple objects have the `id` '
                          'value {!r} in the `{}` array'.format(default_key, '.'.join(map(str, rule_path)))))

        yield new_key, value


def _id_value(key, value, rule):
    # If it is an array of objects, get the `id` value to apply the identifier merge strategy.
    # https://standard.open-contracting.org/latest/en/schema/merging/#identifier-merge
    if 'id' in value:
        id_value = value['id']
        identifier = id_value
    # If the object contained no top-level `id` value, set a unique value.
    else:
        id_value = None
        identifier = str(uuid.uuid1(1))  # use 1 instead of MAC address

    # Calculate the default key for the warning.
    default_key = IdValue(identifier)

    if rule == MergeStrategy.APPEND:
        if 'id' in value:
            new_key = IdValue(str(uuid.uuid1(1)))
        else:  # avoid creating an extra UUID
            new_key = default_key
    elif rule == MergeStrategy.MERGE_BY_POSITION:
        new_key = IdValue(key)
    else:
        new_key = default_key

    # Save the original value. (If the value is an integer, this avoids coercing it to a string.)
    new_key.original_value = id_value

    return new_key, default_key


def unflatten(flattened):
    """
    Unflattens a flattened object into a JSON object.
    """
    unflattened = {}

    identifiers = {}

    for key in flattened:
        current_node = unflattened

        for end, part in enumerate(key, 1):
            # If this is a path to an item of an array.
            # See https://standard.open-contracting.org/1.1/en/schema/merging/#identifier-merge
            if type(part) is IdValue:
                path = key[:end]

                # If the `id` of an object in the array matches, change into it.
                id_path = path + (part.identifier,)
                if id_path in identifiers:
                    current_node = identifiers[id_path]

                # Otherwise, append a new object, and change into it.
                else:
                    new_node = IdDict()
                    new_node.identifier = part.identifier

                    # If the original object had an `id` value, set it.
                    if part.original_value is not None:
                        new_node['id'] = part.original_value

                    # Cache which identifiers appear in which arrays.
                    identifiers[path + (new_node.identifier,)] = new_node

                    current_node.append(new_node)
                    current_node = new_node

                continue

            if not isinstance(current_node, dict):
                message = 'An earlier release had the literal {!r} for /{}, but the current release has an object with a {!r} key'  # noqa: E501
                raise InconsistentTypeError(message.format(current_node, '/'.join(key[:end - 1]), part))

            # Otherwise, this is a path to a property of an object. If this is a path to a node we visited before,
            # change into it. If it's an `id` field, it's already been set to its original value.
            if part in current_node:
                current_node = current_node[part]
                continue

            if end < len(key):
                # If the path is to a new array, start a new array, and change into it.
                if type(key[end]) is IdValue:
                    current_node[part] = []
                # If the path is to a new object, start a new object, and change into it.
                else:
                    current_node[part] = {}

                current_node = current_node[part]
                continue

            # If this is a full path, copy the data, omitting null'ed fields.
            if flattened[key] is not None:
                current_node[part] = flattened[key]

    return unflattened



from ocdsmerge.flatten import flatten, unflatten
from ocdsmerge.rules import get_merge_rules
from ocdsmerge.util import sorted_releases


class Merger:
    def __init__(self, schema=None, merge_rules=None, rule_overrides=None):
        """
        Initializes a reusable ``Merger`` instance for creating merged releases.

        :param schema: the release schema (if not provided, will default to the latest version of OCDS)
        :param dict merge_rules: the merge rules (if not provided, will determine the rules from the ``schema``)
        :param dict rule_overrides: any rule overrides, in which keys are field paths as tuples, and values are either
            ``ocdsmerge.APPEND`` or ``ocdsmerge.MERGE_BY_POSITION``
        :type schema: dict or str
        """
        if merge_rules is None:
            merge_rules = get_merge_rules(schema)
        if rule_overrides is None:
            rule_overrides = {}

        self.merge_rules = merge_rules
        self.rule_overrides = rule_overrides

    def create_compiled_release(self, releases):
        """
        Merges a list of releases into a compiled release.
        """
        return self._create_merged_release(CompiledRelease, releases)

    def create_versioned_release(self, releases):
        """
        Merges a list of releases into a versioned release.
        """
        return self._create_merged_release(VersionedRelease, releases)

    def _create_merged_release(self, cls, releases):
        merged_release = cls(merge_rules=self.merge_rules, rule_overrides=self.rule_overrides)
        merged_release.extend(releases)
        return merged_release.asdict()


class MergedRelease:
    """
    Whether the class is for merging versioned releases.
    """
    versioned = None

    def __init__(self, data=None, schema=None, merge_rules=None, rule_overrides=None):
        """
        Initializes a merged release.

        :param dict data: the latest copy of the merged release, if any
        :param schema: the release schema (if not provided, will default to the latest version of OCDS)
        :param dict merge_rules: the merge rules (if not provided, will determine the rules from the ``schema``)
        :param dict rule_overrides: any rule overrides, in which keys are field paths as tuples, and values are either
            ``ocdsmerge.APPEND`` or ``ocdsmerge.MERGE_BY_POSITION``
        :type schema: dict or str
        """
        if merge_rules is None:
            merge_rules = get_merge_rules(schema)
        if rule_overrides is None:
            rule_overrides = {}

        self.merge_rules = merge_rules
        self.rule_overrides = rule_overrides

        if data is None:
            self.data = {}
        else:
            self.data = flatten(data, self.merge_rules, self.rule_overrides, flattened={}, versioned=self.versioned)

    def asdict(self):
        """
        Returns the merged release as a dictionary.
        """
        return unflatten(self.data)

    def extend(self, releases):
        """
        Sorts and merges many releases into the merged release.
        """
        for release in sorted_releases(releases):
            self.append(release)

    def append(self, release):
        """
        Merges one release into the merged release.
        """
        release = release.copy()

        # Store the values of fields that set "omitWhenMerged": true.
        ocid = release.get('ocid')
        release_id = release.get('id')
        date = release.get('date')
        # Prior to OCDS 1.1.4, `tag` didn't set "omitWhenMerged": true.
        tag = release.pop('tag', None)

        flat = flatten(release, self.merge_rules, self.rule_overrides, flattened={})
        self.flat_append(flat, ocid, release_id, date, tag)

    def flat_append(self, flat, ocid, release_id, date, tag):
        raise NotImplementedError('subclasses must implement flat_append()')


class CompiledRelease(MergedRelease):
    versioned = False

    def __init__(self, data=None, **kwargs):
        super().__init__(data, **kwargs)
        self.data[('tag',)] = ['compiled']

    def flat_append(self, flat, ocid, release_id, date, tag):
        # Add an `id` and `date`.
        self.data[('id',)] = '{}-{}'.format(ocid, date)
        self.data[('date',)] = date
        # In OCDS 1.0, `ocid` incorrectly sets "mergeStrategy": "ocdsOmit".
        self.data[('ocid',)] = ocid

        self.data.update(flat)


class VersionedRelease(MergedRelease):
    versioned = True

    def flat_append(self, flat, ocid, release_id, date, tag):
        # Don't version the OCID.
        flat.pop(('ocid',), None)
        self.data[('ocid',)] = ocid

        for key, value in flat.items():
            # If key is not versioned, continue. If the value is unchanged, don't add it to the history.
            if key in self.data and (type(self.data[key]) is not list or value == self.data[key][-1]['value']):
                continue

            self.data.setdefault(key, []).append({
                'releaseID': release_id,
                'releaseDate': date,
                'releaseTag': tag,
                'value': value,
            })
